# -*- coding: utf-8 -*-
"""
Created on Thu Nov  7 15:31:13 2024

@author: mstep
"""

#!/usr/bin/env python
# coding: utf-8

# %%

import math
import time
import random
import tkinter as tk

num_calls = 0

def make_distances(num_points, points):
    '''
    Make a two-dimensional array giving distances between points
    where distances[i][j] gives the distance from points[i] to points[j].
    '''
    distances = [[0 for i in range(num_points)] for j in range(num_points)]
    for i in range(num_points):
        for j in range(num_points):
            distances[i][j] = math.dist(points[i], points[j])
    return distances

def tour_length(num_points, tour, distances):
    total_length = 0
    for index in range(num_points):
        i = tour[index]
        j = tour[(index + 1) % num_points]
        total_length += distances[i][j]
    return total_length

def make_random_tour_generator(num_points, distances, **kwargs):
    '''Return random tours.'''
    global num_calls
    num_calls = 1

    # Make a list holding the point indices in order so we can shuffle it.
    tour = list(range(num_points))

    # Perform the trials.
    num_trials = kwargs['num_trials']
    for _ in range(num_trials):
        # Pick a random tour and yield it.
        random.shuffle(tour)
        yield tour_length(num_points, tour, distances), tour

def make_exhaustive_tour_generator(num_points, distances, **_):
    global num_calls
    num_calls = 0

    # Prepare to call generate_exhaustive_tours.
    is_used = [False for i in range(num_points)]

    # Arbitrarily start at point 0.
    is_used[0] = True
    current_tour = [0]

    # Solve recursively.
    for length, tour in generate_exhaustive_tours(
            num_points, distances, is_used, current_tour):
        yield length, tour

def generate_exhaustive_tours(num_points, distances, is_used, current_tour):
    '''Generate all possible tours.'''
    global num_calls
    num_calls += 1

    # See if we have a full tour.
    if len(current_tour) == num_points:
        # We have a full tour. Yield it.
        length = tour_length(num_points, current_tour, distances)
        yield length, current_tour
    else:
        # Pick the next point to add to the tour.
        for i in range(num_points):
            # Try to add points[i].
            if not is_used[i]:
                # Add it.
                is_used[i] = True
                current_tour.append(i)

                # Recursively complete the tour.
                # Repeat until the nested generator runs dry.
                for length, tour in \
                        generate_exhaustive_tours(
                            num_points, distances, is_used, current_tour):
                    yield length, tour

                # Remove points[i] so we can use it again later.
                is_used[i] = False
                current_tour.pop()

def make_b_and_b_tour_generator(num_points, distances, **_):
    '''Use branch and bound to find tours.'''
    global num_calls
    num_calls = 0

    # Prepare to call generate_b_and_b_tours.
    is_used = [False for i in range(num_points)]

    # Arbitrarily start at point 0.
    is_used[0] = True
    current_tour = [0]
    current_length = 0

    # Solve recursively.
    best_length = math.inf
    for length, tour in generate_b_and_b_tours(
            num_points, distances, is_used, current_tour, current_length,
            best_length):
        yield length, tour

def generate_b_and_b_tours(num_points, distances, is_used,
        current_tour, current_length, best_length):
    '''Generate all possible tours.'''
    global num_calls
    num_calls += 1

    # See if we have a full tour.
    if len(current_tour) == num_points:
        # We have a full tour.
        # Add the distance from the last point back to the first.
        current_length += distances[current_tour[-1]][current_tour[0]]

        # Yield it.
        yield current_length, current_tour
    else:
        # Pick the next point to add to the tour.
        for i in range(num_points):
            # Try to add points[i].
            if not is_used[i]:
                # Get the distance needed to add point i to the tour.
                distance_to_i = distances[current_tour[-1]][i]
                current_length += distance_to_i 

                # Add it.
                is_used[i] = True
                current_tour.append(i)

                # Calculate the best possible tour length from here.
                test_length = current_length + \
                    best_possible_length(num_points, distances, is_used)
                if test_length < best_length:
                    # Recursively complete the tour.
                    # Repeat until the nested generator runs dry.
                    for length, tour in generate_b_and_b_tours(
                            num_points, distances, is_used,
                            current_tour, current_length, best_length):
                        # Update the best tour length.
                        if length < best_length:
                            best_length = length

                            # Yield the tour.
                            yield length, tour

                # Remove points[i] so we can use it again later.
                current_length -= distance_to_i
                is_used[i] = False
                current_tour.pop()

def best_possible_length(num_points, distances, is_used):
    '''Return the best possible tour length starting from the current tour.'''
    # Loop through the points.
    total_length = 0
    for i in range(num_points):
        # If point i is unused...
        if not is_used[i]:
            # ...find smallest distance into point i.
            length_to_i = math.inf
            for j in range(num_points):
                if j != i and distances[j][i] < length_to_i:
                    length_to_i = distances[j][i]
            total_length += length_to_i
    return total_length

# In[2]:

class TspApp:
    def __init__(self):
        self.best_tour = None
        self.best_length = math.inf
        self.distances = None
        self.ms = 0
        self.start_time = None
        self.tour_generator = None
        self.tour_number = 0

        self.window = tk.Tk()
        self.window.title('tsp_b_and_b')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('700x500')

        # Left frame.
        frame = tk.Frame(self.window)
        frame.pack(side=tk.LEFT, anchor=tk.N, padx=3, pady=3)

        # # Points.
        LABEL_WIDTH = 14
        ENTRY_WIDTH = 8
        self.num_points_entry, self.num_points_var = \
            self.make_field(frame, LABEL_WIDTH, "# Points:", ENTRY_WIDTH, '6')

        # Buttons.
        BUTTON_WIDTH = 15
        self.generate_button = self.make_button(
            frame, BUTTON_WIDTH, 'Generate', self.generate_click, tk.NORMAL)

        self.num_trials_entry, self.num_trials_var = \
            self.make_field(frame, LABEL_WIDTH, "# Trials:", ENTRY_WIDTH, '100')

        delay_frame = tk.Frame(frame, borderwidth=2, relief=tk.SUNKEN)
        delay_frame.pack(side=tk.TOP, fill=tk.X, pady=5)
        label = tk.Label(delay_frame, text='Delay (ms):')
        label.pack(side=tk.TOP)

        self.ms_scale = tk.Scale(delay_frame, from_=0, to=1000,
            command=self.ms_changed, orient=tk.HORIZONTAL)
        self.ms_scale.pack(side=tk.TOP, fill=tk.X)
        self.ms_scale.set(250)

        self.random_button = self.make_button(
            frame, BUTTON_WIDTH, 'Random', self.random_click, tk.DISABLED)
        self.exhaustive_button = self.make_button(
            frame, BUTTON_WIDTH, 'Exhaustive', self.exhaustive_click,
            tk.DISABLED)
        self.b_and_b_button = self.make_button(
            frame, BUTTON_WIDTH, 'Branch & Bound', self.b_and_b_click,
            tk.DISABLED)

        # Outputs.
        _, self.tour_number_var = self.make_field(
            frame, LABEL_WIDTH, "Tour #:", 10, '', 'readonly')
        _, self.best_length_var = self.make_field(
            frame, LABEL_WIDTH, "Tour Length:", 10, '', 'readonly')
        _, self.elapsed_time_var = self.make_field(
            frame, LABEL_WIDTH, "Time (sec):", 10, '', 'readonly')
        _, self.num_calls_var = self.make_field(
            frame, LABEL_WIDTH, "# Calls:", 10, '', 'readonly')

        # Canvas.
        self.canvas = tk.Canvas(self.window, bg='white', borderwidth=2,
                                relief=tk.SUNKEN)
        self.canvas.pack(side=tk.LEFT, fill=tk.BOTH,
                         expand=True, padx=(0,5), pady=5)

        # Bind the Return key.
        self.num_points_entry.bind("<Return>", lambda e: self.generate_click())

        self.num_points_entry.focus_force()
        self.window.mainloop()

    def make_field(self, parent, label_width, label_text,
                   entry_width, entry_text, entry_state=tk.NORMAL):
        '''
        Make a Label and associated Entry in columns 0 and 1.
        Return the Entry and its variable.
        '''
        frame = tk.Frame(parent)
        frame.pack(side=tk.TOP)

        label = tk.Label(frame, text=label_text, width=label_width,
            anchor=tk.W)
        label.pack(side=tk.LEFT, padx=3, pady=3)

        var = tk.StringVar(parent, value=entry_text)
        entry = tk.Entry(frame, textvariable=var, width=entry_width,
            justify=tk.RIGHT, state=entry_state)
        entry.pack(side=tk.LEFT, padx=3, pady=5)
        return entry, var

    def make_button(self, parent, button_width, text, command, state):
        '''Make a Button in this row spanning columns 0 and 1.'''
        button = tk.Button(parent, text=text, width=button_width,
                           command=command, state=state)
        button.pack(side=tk.TOP, padx=3, pady=3)
        return button

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def disable_solver_buttons(self):
        '''Disable the generate and solver buttons.'''
        self.generate_button.configure(state=tk.DISABLED)
        self.random_button.configure(state=tk.DISABLED)
        self.exhaustive_button.configure(state=tk.DISABLED)
        self.b_and_b_button.configure(state=tk.DISABLED)

    def enable_solver_buttons(self):
        '''Enable appropriate generate and solver buttons.'''
        self.generate_button.configure(state=tk.NORMAL)
        self.random_button.configure(state=tk.NORMAL)
        if self.num_points <= 12:
            # In one set of tests:
            #   10 points:   11.78 seconds
            #   11 points:  159.32 seconds =  2 minutes 39 seconds
            #   12 points: 2103.36 seconds = 35 minutes  3 seconds
            self.exhaustive_button.configure(state=tk.NORMAL)
        else:
            self.exhaustive_button.configure(state=tk.DISABLED)
        if self.num_points <= 20:
            # In one set of tests:
            #   13 points:   7.44,   4.71,   3.64 seconds
            #   14 points:  10.61,  13.01,  83.96 seconds
            #   15 points:  29.64, 316.24, 126.10, seconds
            self.b_and_b_button.configure(state=tk.NORMAL)
        else:
            self.b_and_b_button.configure(state=tk.DISABLED)

    def generate_click(self):
        # Start with no tour.
        self.best_tour = None
        self.best_length = math.inf

        # Pick random points.
        wid = self.canvas.winfo_width()
        hgt = self.canvas.winfo_height()
        MARGIN = 20
        self.num_points = int(self.num_points_var.get())
        self.points = []
        for _ in range(self.num_points):
            x = random.randint(MARGIN, wid - MARGIN)
            y = random.randint(MARGIN, hgt - MARGIN)
            self.points.append((x, y))

        # Find distances between all pairs of points.
        self.distances = make_distances(self.num_points, self.points)

        # Enable solver and output buttons.
        self.enable_solver_buttons()
        self.tour_number_var.set('')
        self.best_length_var.set('')
        self.elapsed_time_var.set('')
        self.num_calls_var.set('')

        self.draw()

    def ms_changed(self, _):
        self.ms = self.ms_scale.get()

    def random_click(self):
        '''Solve by using random orderings of the points.'''
        print('Random:     ', end='')
        self.solve(make_random_tour_generator)

    def exhaustive_click(self):
        '''Solve by using random orderings of the points.'''
        print('Exhaustive: ', end='')
        self.solve(make_exhaustive_tour_generator)

    def b_and_b_click(self):
        '''Solve by using random orderings of the points.'''
        print('B & B:      ', end='')
        self.solve(make_b_and_b_tour_generator)

    def solve(self, make_tour_generator):
        '''Solve by using the given generator.'''
        # Reset any previous tour.
        self.best_tour = None
        self.best_length = math.inf

        # Disable the solver buttons.
        self.disable_solver_buttons()
        self.best_length_var.set(f'{self.best_length:.0f}')
        self.elapsed_time_var.set('')
        self.num_calls_var.set('')

        # Make a tour generator.
        num_trials = int(self.num_trials_var.get())
        self.tour_generator = make_tour_generator(self.num_points,
            self.distances, num_trials=num_trials)
        self.tour_number = 0

        # Start generating tours.
        self.window.update()
        self.start_time = time.time()
        self.check_next_tour()

    def check_next_tour(self):
        global num_calls

        try:
            # Get the next tour. This throws
            # StopIteration if the generator is empty.
            test_length, test_tour = next(self.tour_generator)

            # Display the tour number.
            self.tour_number += 1
            self.tour_number_var.set(f'{self.tour_number}')

            # Draw the test tour (if we're not going as fast as possible).
            if self.ms > 0:
                self.draw(test_tour)

            # See if this is an improvement.
            if test_length < self.best_length:
                # Save the improved tour.
                self.best_length = test_length
                self.best_tour = test_tour.copy()

                # Display the improved length.
                self.best_length_var.set(f'{self.best_length:.0f}')

            # Pause and repeat.
            self.window.after(self.ms, self.check_next_tour)

        except StopIteration:
            # We've finished the desired number of trials.
            stop_time = time.time()
            self.best_length_var.set(f'{self.best_length:.0f}')
            self.elapsed_time_var.set(f'{stop_time - self.start_time:.2f}')
            self.num_calls_var.set(f'{num_calls}')

            # Print results.
            print(f'Length: {self.best_length:>6.0f}, ' + \
                  f'Time: {stop_time - self.start_time:.2f}, ' + \
                  f'Calls: {num_calls:>6d}')

            # Re-enable the solver buttons.
            self.enable_solver_buttons()

            # Draw the best tour.
            self.draw(self.best_tour)

    def draw(self, tour=None):
        '''Draw the points and tours.'''
        # Delete any previous drawing objects.
        self.canvas.delete(tk.ALL)

        # Draw the points.
        RADIUS = 5
        COLOR = 'blue'
        for point in self.points:
            self.canvas.create_oval(
                point[0] - RADIUS, point[1] - RADIUS,
                point[0] + RADIUS, point[1] + RADIUS,
                fill=COLOR)

        # Draw the tour.
        if tour is not None:
            tour_points = []
            for index in tour:
                tour_points.append(self.points[index])
            self.canvas.create_polygon(tour_points, outline='blue', fill='')

# In[3]:

TspApp()
print('Done')


# In[ ]:
